package data.scripts.ai;

import com.fs.starfarer.api.Global;
import com.fs.starfarer.api.combat.CombatEntityAPI;
import com.fs.starfarer.api.combat.GuidedMissileAI;
import com.fs.starfarer.api.combat.MissileAIPlugin;
import com.fs.starfarer.api.combat.MissileAPI;
import com.fs.starfarer.api.combat.ShipAPI;
import com.fs.starfarer.api.combat.ShipAPI.HullSize;
import com.fs.starfarer.api.combat.ShipCommand;
import com.fs.starfarer.api.util.IntervalUtil;
import java.util.Collections;
import java.util.List;
import java.util.ListIterator;

import org.lwjgl.util.vector.Vector2f;

import org.lazywizard.lazylib.CollectionUtils;
import org.lazywizard.lazylib.MathUtils;
import org.lazywizard.lazylib.VectorUtils;
import org.lazywizard.lazylib.combat.AIUtils;
import org.lazywizard.lazylib.combat.CombatUtils;

public class ParabolaAI implements MissileAIPlugin, GuidedMissileAI {

    private final MissileAPI missile;
    private CombatEntityAPI target;
    private float offTarget;
    private float baseOffTargetFraction;
    private float offTargetAngle;
    private IntervalUtil timer = new IntervalUtil(0.1f, 0.2f);

    private final static float ACCELERATION = 100f;
    private final static float MAX_SPEED = 300f;

    private static ShipAPI findBestTarget(MissileAPI missile) {
        ShipAPI source = missile.getSource();
        if (source != null && source.getShipTarget() != null && !source.getShipTarget().isHulk()) {
            return source.getShipTarget();
        }

        ShipAPI closest = null;
        float distance, closestDistance = Float.MAX_VALUE;

        for (ShipAPI tmp : AIUtils.getEnemiesOnMap(missile)) {
            if (tmp.isFighter() || tmp.isDrone() || tmp.getHullSize() == HullSize.FIGHTER) {
                continue;
            }
            distance = MathUtils.getDistance(tmp, missile.getLocation());
            if (distance < closestDistance) {
                closest = tmp;
                closestDistance = distance;
            }
        }

        return closest;
    }

    public ParabolaAI(MissileAPI missile, ShipAPI launchingShip) {
        this.missile = missile;
        this.offTarget = 40f * (0.5f - (float) Math.random());
        this.baseOffTargetFraction = 3f * (0.5f - (float) Math.random());
        this.offTargetAngle = 360f * (float) Math.random();

        List<ShipAPI> directTargets = CombatUtils.getShipsWithinRange(launchingShip.getMouseTarget(), 100f);
        if (!directTargets.isEmpty()) {
            Collections.sort(directTargets, new CollectionUtils.SortEntitiesByDistance(launchingShip.getMouseTarget()));
            ListIterator<ShipAPI> iter = directTargets.listIterator();
            while (iter.hasNext()) {
                ShipAPI tmp = iter.next();
                if (tmp.isFighter() || tmp.isDrone() || tmp.getHullSize() == HullSize.FIGHTER) {
                    continue;
                }
                if (!tmp.isHulk() && tmp.getOwner() != launchingShip.getOwner()) {
                    setTarget(tmp);
                    break;
                }
            }
        }

        if (target == null) {
            target = launchingShip.getShipTarget();
        }

        if (target == null) {
            setTarget(findBestTarget(missile));
        }
    }

    @Override
    public void advance(float amount) {
        timer.advance(amount);

        if (missile.isFading() || missile.isFizzling()) {
            return;
        }

        // If our current target is lost, assign a new one
        if (target == null || (target instanceof ShipAPI && ((ShipAPI) target).isHulk()) || (missile.getOwner() == target.getOwner()) || !Global.getCombatEngine().isEntityInPlay(target)) {
            setTarget(findBestTarget(missile));
            missile.giveCommand(ShipCommand.ACCELERATE);
            return;
        }

        if (timer.intervalElapsed()) {
            offTarget = (offTarget > 0f ? offTarget - 0.5f : offTarget + 0.5f);
            baseOffTargetFraction *= 0.99f;
        }

        float offTargetBy = (0f + (offTarget + (baseOffTargetFraction * target.getCollisionRadius())));
        Vector2f offset = VectorUtils.rotate(new Vector2f(offTargetBy, 0f), offTargetAngle, new Vector2f());
        Vector2f targetLoc = new Vector2f(target.getLocation());
        Vector2f.add(offset, targetLoc, targetLoc);

        float distance = MathUtils.getDistance(targetLoc, missile.getLocation());
        float acceleration = ACCELERATION;
        float maxSpeed = MAX_SPEED;
        if (missile.getSource() != null) {
            acceleration = missile.getSource().getMutableStats().getMissileAccelerationBonus().computeEffective(acceleration);
            maxSpeed = missile.getSource().getMutableStats().getMissileMaxSpeedBonus().computeEffective(maxSpeed);
        }
        Vector2f guidedTarget = intercept(missile.getLocation(), missile.getVelocity().length(), acceleration, maxSpeed, targetLoc, target.getVelocity());
        if (guidedTarget == null) {
            Vector2f projection = new Vector2f(target.getVelocity());
            float scalar = distance / (missile.getVelocity().length() + 1f);
            projection.scale(scalar);
            guidedTarget = Vector2f.add(targetLoc, projection, null);
        }

        float velocityFacing = VectorUtils.getFacing(missile.getVelocity());
        float absoluteDistance = MathUtils.getShortestRotation(velocityFacing, VectorUtils.getAngle(missile.getLocation(), guidedTarget));
        float angularDistance = MathUtils.getShortestRotation(missile.getFacing(), VectorUtils.getAngle(missile.getLocation(), guidedTarget));
        float compensationDifference = MathUtils.getShortestRotation(angularDistance, absoluteDistance);
        if (Math.abs(compensationDifference) <= 75f) {
            angularDistance += 0.5f * compensationDifference;
        }
        float absDVel = Math.abs(absoluteDistance);
        float absDAng = Math.abs(angularDistance);

        missile.giveCommand(angularDistance < 0 ? ShipCommand.TURN_RIGHT : ShipCommand.TURN_LEFT);

        if (absDVel >= 135f || ((absDVel <= 90f && (absDVel <= 45f || distance >= 500f)) || missile.getVelocity().length() <= maxSpeed * 0.75f)) {
            missile.giveCommand(ShipCommand.ACCELERATE);
        }

        // Course correction
        if (absDAng < 5) {
            float MFlightAng = VectorUtils.getAngle(new Vector2f(0, 0), missile.getVelocity());
            float MFlightCC = MathUtils.getShortestRotation(missile.getFacing(), MFlightAng);
            if (Math.abs(MFlightCC) > 20) {
                missile.giveCommand(MFlightCC < 0 ? ShipCommand.STRAFE_LEFT : ShipCommand.STRAFE_RIGHT);
            }
        }

        if (absDAng < 0.4) {
            missile.setAngularVelocity(0);
        }
    }

    @Override
    public CombatEntityAPI getTarget() {
        return target;
    }

    @Override
    public final void setTarget(CombatEntityAPI target) {
        this.target = target;
    }

    private static Vector2f intercept(Vector2f point, float speed, float acceleration, float maxspeed, Vector2f target, Vector2f targetVel) {
        Vector2f difference = new Vector2f(target.x - point.x, target.y - point.y);

        float s = speed;
        float a = acceleration / 2f;
        float b = speed;
        float c = difference.length();
        Vector2f solutionSet = quad(a, b, c);
        if (solutionSet != null) {
            float t = Math.min(solutionSet.x, solutionSet.y);
            if (t < 0) {
                t = Math.max(solutionSet.x, solutionSet.y);
            }
            if (t > 0) {
                s = acceleration * t;
                s = s / 2f + speed;
                s = Math.min(s, maxspeed);
            }
        }

        a = targetVel.x * targetVel.x + targetVel.y * targetVel.y - s * s;
        b = 2 * (targetVel.x * difference.x + targetVel.y * difference.y);
        c = difference.x * difference.x + difference.y * difference.y;

        solutionSet = quad(a, b, c);

        Vector2f intercept = null;
        if (solutionSet != null) {
            float bestFit = Math.min(solutionSet.x, solutionSet.y);
            if (bestFit < 0) {
                bestFit = Math.max(solutionSet.x, solutionSet.y);
            }
            if (bestFit > 0) {
                intercept = new Vector2f(target.x + targetVel.x * bestFit, target.y + targetVel.y * bestFit);
            }
        }

        return intercept;
    }

    private static Vector2f quad(float a, float b, float c) {
        Vector2f solution = null;
        if (Float.compare(Math.abs(a), 0) == 0) {
            if (Float.compare(Math.abs(b), 0) == 0) {
                solution = (Float.compare(Math.abs(c), 0) == 0) ? new Vector2f(0, 0) : null;
            } else {
                solution = new Vector2f(-c / b, -c / b);
            }
        } else {
            float d = b * b - 4 * a * c;
            if (d >= 0) {
                d = (float) Math.sqrt(d);
                a = 2 * a;
                solution = new Vector2f((-b - d) / a, (-b + d) / a);
            }
        }
        return solution;
    }
}
